@@ -151,11 +151,11 @@ static char *safe_env_lst[] =
     "USER_NAME",
     "TZ",
 // #if !DO_CHROOT
-    "SCRIPT_NAME",
-    "PATH_INFO",
-    "PATH_TRANSLATED",
-    "DOCUMENT_ROOT",
-    "SCRIPT_FILENAME",
+//    "SCRIPT_NAME",
+//    "PATH_INFO",
+//    "PATH_TRANSLATED",
+//    "DOCUMENT_ROOT",
+//    "SCRIPT_FILENAME",
 // #endif
     NULL
 };
@@ -480,7 +480,7 @@ void read_config_file(char* sboxconf_path){
 	return;
 }
 
-// Maxmimum string size for an environment value:
+// Maxmimum string size for an environment name or value:
 #define MAXENV 1024
 
 // Get the HTTP_HOST environment variable value
@@ -492,6 +492,7 @@ char* get_http_host(char** e){
 		delimiter = (char*) strchr(name,'=');
 		if (delimiter == NULL) continue;        // If we don't find the "=" delimiter, then we didn't find anything...
 		if (delimiter - name >= MAXENV) continue;       // If our envvar name is too big, then it's as if we didn't find anything
+		if (strlen(delimiter) > MAXENV) continue;
 		strncpy(envvar_name,name,delimiter-name);        // Copy the envvar name in a buffer, to check for it later
 		envvar_name[delimiter-name]='\0';                // Make sure it terminates by a delimiting zero
 		delimiter++;                            // Point just right after the = sign
@@ -513,6 +514,7 @@ void print_the_env(char** e){
 		delimiter = (char*) strchr(name,'=');
 		if (delimiter == NULL) continue;
 		if (delimiter - name >= MAXENV) continue;
+		if (strlen(delimiter) > MAXENV) continue;
 		strncpy(nm,name,delimiter-name);
 		nm[delimiter-name]='\0';
 		log_error("%s => %s\n",nm,++delimiter);
@@ -543,6 +545,7 @@ void readEnvSboxConfig(char** e){
 		delimiter = (char*) strchr(name,'=');
 		if (delimiter == 0) continue;
 		if (delimiter - name >= MAXENV) continue;
+		if (strlen(delimiter) > MAXENV) continue;
 		strncpy(nm,name,delimiter-name);
 		nm[delimiter-name]='\0';
 		delimiter++;
@@ -711,9 +714,6 @@ int main (int argc, char* argv[], char** e) {
 	printf("limit_nofile_soft: %d\n",sbox_config.limit_nofile_soft);
 #endif
 
-	if(sbox_config.do_log_envvars){
-		print_the_env(e);
-	}
 	open_log();
 
 	if(sbox_config.do_log_envvars){
@@ -805,9 +805,6 @@ int main (int argc, char* argv[], char** e) {
 		fatal_error("couldn't setuid() to %d (%s)\n",getuid(),strerror(errno));
 	}
   
-	if(strlen(sbox_config.log_file) > 0){
-		log_error("executing %s for uid %d, gid %d\n",target,getuid(),getgid());
-	}
 
 
 	/* fix up argv[] array and command depending on the extension */
@@ -815,7 +812,9 @@ int main (int argc, char* argv[], char** e) {
 	if(ext == 0){
 		new_cmd_interp = cmd;
 		argv[0] = cmd;
-		log_error("No extension found: starting as CGI\n");
+		if(strlen(sbox_config.log_file) > 0){
+			log_error("No extension found: starting as CGI: executing %s for uid %d, gid %d\n",target,getuid(),getgid());
+		}
 	}else{
 		if( strcmp(ext,".php") == 0){
 			new_cmd_interp = sbox_config.php_path;
@@ -831,18 +830,20 @@ int main (int argc, char* argv[], char** e) {
 		}
 		argv[0] = new_cmd_interp;
 		argv[1] = cmd;
-		log_error("Extension is: %s, will use %s as interpreter\n",ext,new_cmd_interp);
+		log_error("Found %s for %s: %s %s (%s for uid %d and gid %d)\n",ext,target,new_cmd_interp,argv[1],cleanenv[0],getuid(),getgid());
 	}
 
 	/* chdir() to new directory */
 	chdir_fix(newroot,target);
-	if(strlen(sbox_config.log_file) > 0){
+	if(strlen(sbox_config.log_file) > 0 && ext == 0){
 		log_error("Going to try this: %s , %s, %s\n", new_cmd_interp, argv[1], cleanenv[0]);
 	}
 
-
-//if (log && log != stderr)
-//    fclose(log);
+	/* close log file so we don't leak file descriptor */
+	if (log && log != stderr) {
+		fp = log; log = NULL;
+		fclose(fp);
+	}
 
 	/* execute the target */
 	return_int = execve(new_cmd_interp,argv,cleanenv);
@@ -912,7 +913,7 @@ static void write_log(const char* fmt, va_list ap) {
 	timestr = ctime(&timevar);
 	timestr[strlen(timestr)-1]='\0';
   
-	fprintf(log,"[%s] sbox: ",timestr);
+	fprintf(log,"[%s] sbox[%d]: ",timestr,getpid());
 	vfprintf(log, fmt, ap);
 	fflush(log);
 
@@ -1289,78 +1290,79 @@ char* shave(char* newroot,
 }
 
 /* modify the environment in the following ways: 
- * DOCUMENT_ROOT => shave(DOCUMENT_ROOT)
- * PATH_INFO => path_info
- * PATH_TRANSLATED => "DOCUMENT_ROOT/path_info"
- * SCRIPT_FILENAME => shave(target)
- * SCRIPT_NAME => "$SCRIPT_NAME/shave(target)"
+ * DOCUMENT_ROOT => shave(newroot,root)
+ * SCRIPT_NAME => shave(root,target)
+ * SCRIPT_FILENAME => shave(newroot,target)
+ * PATH_INFO => shave(SCRIPT_NAME,env(path_info))
+ * PATH_TRANSLATED => SCRIPT_FILENAME,env(path_info)
  */
 static void tweak_environment(char* newroot,
 			      char* root,
 			      char* target,
 			      char* additional_info) {
-	char *s,*t,*path_info,*script_name;
+	char *s,*t,*path_info,*path_translated,*script_name;
 	int r,max;
 
-	/* the SCRIPT_NAME needs to be the URL that we invoke to get at the
-	 script again.  */
-	path_info = (char*) getenv("PATH_INFO");
-	script_name = (char*) getenv("SCRIPT_NAME");
-	if (script_name == NULL)
-		fatal_error("expect SCRIPT_NAME environment variable to be set, but null found\n");
-
-	/* set the script name -- it should be existing SCRIPT_NAME concatenated
-	   to the path info, less the additional info.  SCRIPT_NAME can be used to
-	   reinvoke the script in virtual name space*/
-	max = strlen(script_name) +  strlen(path_info) -  strlen(additional_info) +  1;
-
-	s = (char*) calloc(sizeof(char),max);
-	if (s == NULL)
-		fatal_error("failed to allocate memory for SCRIPT_NAME: %s\n",strerror(errno));
-
-	strncpy(s,script_name,max);
-	strncat(s,path_info,strlen(path_info)-strlen(additional_info));
-	s[max-1] = '\0';
-
-	r = set_env("SCRIPT_NAME",s,1);
+	/* per rfc3875:  The SCRIPT_NAME variable MUST be set to a URI path
+	   (not URL-encoded) which could identify the CGI script ...
+	   No PATH_INFO segment is included in the SCRIPT_NAME value. */
+	script_name = shave(root,target);
+	r = set_env("SCRIPT_NAME",script_name,1);
 	if (r < 0) 
 		fatal_error("no room for SCRIPT_NAME environment variable");
 
-	free((void*)s);
-
-	if (*additional_info) {
-		/* set the path info */
-		r = set_env("PATH_INFO",additional_info,1);
-		if (r < 0) 
-			fatal_error("no room for PATH_INFO environment variable");
-
-		t = shave(newroot,root);
-
-		max = strlen(t)+strlen(additional_info)+1;
-		if (!(s = (char*) calloc(sizeof(char),max)))
-			fatal_error("unable to allocate string for path translated: %s\n",strerror(errno));
-
-		/* set the path translated */
-		strncpy(s,t,max);
-		if (t[strlen(t)-1] == '/')
-			strncat(s,additional_info+1,strlen(additional_info)-1);
-		else
-			strncat(s,additional_info,strlen(additional_info));
-		s[max-1] = '\0';
-		r = set_env("PATH_TRANSLATED",s,1);
-		if (r < 0) 
-			fatal_error("no room for PATH_TRANSLATED environment variable");
-
-		free((void*)s);
-	}
-
-	/* Shave the chroot directory off the target.  Will error out
+	/* Shave the chroot directory off the target.  Will cause an error
 	   if the target is not contained within the new root. */
 	s = shave(newroot,target);
 	r = set_env("SCRIPT_FILENAME",s,1);
 	if (r < 0) 
 		fatal_error("no room for SCRIPT_FILENAME environment variable");
 
+	/* PATH_INFO Contains any client-provided pathname information trailing
+	   the actual script filename but preceding the query string, if available.
+	   It does not include the script name itsself (SCRIPT_NAME). */
+	s = (char*) getenv("PATH_INFO");
+	path_info = shave(script_name,s);
+//	r = set_env("PATH_INFO",(strlen(path_info) ? path_info : "/"),1);
+//	if (r < 0)
+//		fatal_error("no room for PATH_INFO environment variable");
+
+	/* PATH_TRANSLATED "is derived by taking the PATH_INFO value, parsing it
+	   as a local URI in its own right, and performing any virtual-to-physical
+	   translation appropriate to map it onto the server's document repository
+	   structure."  So says rfc3875.  That would make it roughly
+	   DOCUMENT_ROOT/uri_encode(PATH_INFO) .. which seems fairly useless.
+	   Additionally, "If PATH_INFO is NULL, then the PATH_TRANSLATED variable
+	   MUST be set to NULL (or unset)."
+	   */
+	if (strlen(path_info)) {
+		/* this is DOCUMENT_ROOT */
+		s = shave(newroot,root);
+
+		/* PATH_TRANSLATED should already be properly URI-encoded */
+		path_translated = (char*) getenv("PATH_TRANSLATED");
+		t = shave(target,path_translated);
+
+		//max = strlen(s)+((t[0] != '/')?1:0)+ strlen(t)+1 +strlen(script_name);
+		max = strlen( shave(newroot,target) ) + strlen(path_info) + 1;
+		if (!(path_translated = (char*) calloc(sizeof(char),max)))
+			fatal_error("unable to allocate string for path translated: %s\n",strerror(errno));
+
+		sprintf(path_translated,"%s%s",shave(newroot,target),path_info);
+
+		r = set_env("PATH_TRANSLATED",path_translated,1);
+		if (r < 0)
+			fatal_error("no room for PATH_TRANSLATED environment variable");
+
+		//script_name = shave(root,target);
+		max = strlen( shave(root,target) ) + strlen(path_info) + 1;
+		if (!(s = (char*) calloc(sizeof(char),max)))
+			fatal_error("unable to allocate string for path_info: %s\n",strerror(errno));
+		sprintf(s,"%s%s", shave(root,target),path_info);
+		r = set_env("PATH_INFO",s,1);
+		free((void*)s);
+	}
+
 	s = shave(newroot,root);
 	r = set_env("DOCUMENT_ROOT",s,1);
 	if (r < 0) 
@@ -1414,20 +1416,32 @@ static int set_env(const char *name, const char *value, int overwrite) {
    part of the Apache distribution */
 void clean_env(){
 	char pathbuf[NAME_MAX];
+	char *name,*delimiter;
 	char **ep;
 	int cidx = 0;
 	int idx;
   
-
 	/* initialize cleanenv global if necessary */
 	if (cleanenv == NULL) {
 		if ((cleanenv = (char **)calloc(AP_ENVBUF, sizeof(char *))) == NULL)
 			fatal_error("failed to malloc env mem\n");
 	}
 
-	/* position to first NULL entry in cleanenv */
-	for (cidx = 0; cleanenv[cidx] && cidx < AP_ENVBUF; cidx++)
-		;
+	/* check cleanenv name and value lengths, and position to first NULL entry */
+	for (cidx = 0; cleanenv[cidx] && cidx < AP_ENVBUF; cidx++) {
+		name = cleanenv[cidx];
+		delimiter = (char*) strchr(name,'=');
+		if (delimiter == 0)
+			fatal_error("malformed cleanenv entry %d, not in name=value format\n",cidx);
+		if (delimiter - name >= MAXENV) {
+			name[MAXENV] = '\0';
+			fatal_error("env name too long: %s\n",name);
+		}
+		if (strlen(delimiter) > MAXENV) {
+			delimiter[MAXENV] = '\0';
+			fatal_error("env value too long: %s = %s\n",name,++delimiter);
+		}
+	}
 
 	/* copy the environment */
 	for (ep = environ; *ep && cidx < AP_ENVBUF; ep++) {
